#-*- coding: utf-8 -*-
from socket import error as socket_error
from django.http import BadHeaderError


class WebSocketError(socket_error):
    """
    WebSocket常规错误
    """


class FrameTooLargeException(WebSocketError):
    """
    Raised if a received frame is too large.
    """


class HandshakeError(BadHeaderError):
    """
    握手发生错误了
    """


class UpgradeRequiredError(HandshakeError):
    """
    Raised if protocol must be upgraded.
    """



from django.conf import settings

WS_SETTINGS = {
    "WEBSOCKET_URL": getattr(settings, 'WEBSOCKET_URL', '/ws/'),
    "CHANNEL_MIDDLEWARE": getattr(settings, 'CHANNEL_MIDDLEWARE', 'websocket_channel.middleware.ChannelMiddleware'),
    "OTHER_REQUET": getattr(settings, 'OTHER_REQUET', None),
    "WS_REDIS": getattr(settings, 'WS_CONNECTION', {
        'host': 'localhost',
        'port': 6379,
        'db': 0,
        'password': 'foobared',
    }),
}




# -*- coding: utf-8 -*-
import json
import six

class RedisMessage(six.binary_type):

    def __new__(cls, value):
        if isinstance(value, six.string_types):
            return six.binary_type.__new__(cls, value)
        elif isinstance(value, list):
            if len(value) >= 2 and value[0] == 'message':
                channel = value[1]
                return six.binary_type.__new__(cls, json.dumps({'channel': channel, 'content': value[2]}))
        return None


class RedisStore(object):
    """
    Abstract base class to control publishing and subscription for messages to and from the Redis
    datastore.
    """

    def __init__(self, connection):
        self._connection = connection
        self._publishers = set()

    def publish_message(self, message):

        if not isinstance(message, RedisMessage):
            message = RedisMessage(message)

        for channel in self._publishers:
            self._connection.publish(channel, message)



import json
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from websocket_channel.publisher import RedisPublisher


@csrf_exempt
def test1(request):
    RedisPublisher(request.POST['channel']).publish_message(request.POST['content'])
    return HttpResponse()


def websocket_request(request, recvmsg, subscriber):
    RedisPublisher("1111111111").publish_message("1111111111")


# -*- coding: utf-8 -*-
from .web_socket_message import RedisStore


class RedisSubscriber(RedisStore):

    def __init__(self, connection):
        self._subscription = None
        super(RedisSubscriber, self).__init__(connection)

    def parse_response(self):
        """
        Parse a message response sent by the Redis datastore on a subscribed channel.
        """
        return self._subscription.parse_response()

    def set_pubsub_channels(self, request, channels):
        self._subscription = self._connection.pubsub()
        for channel in channels:
            self._subscription.subscribe(channel)

    def subscribe(self, channel):
        if self._subscription and channel not in self._subscription.channels:
            self._subscription.subscribe(channel)

    def send_persited_messages(self, websocket):
        """
        This method is called immediately after a websocket is openend by the client, so that
        persisted messages can be sent back to the client upon connection.
        """
        for channel in self._subscription.channels:
            message = self._connection.get(channel)
            if message:
                websocket.send(message)

    def get_file_descriptor(self):
        """
        Returns the file descriptor used for passing to the select call when listening
        on the message queue.
        """
        return self._subscription.connection and self._subscription.connection._sock.fileno()

    def release(self):
        """
        New implementation to free up Redis subscriptions when websockets close. This prevents
        memory sap when Redis Output Buffer and Output Lists build when websockets are abandoned.
        """
        if self._subscription and self._subscription.subscribed:
            self._subscription.unsubscribe()
            self._subscription.reset()




#-*- coding: utf-8 -*-
import six
import base64
import select
from hashlib import sha1
from wsgiref import util
from django.core.wsgi import get_wsgi_application
from django.core.servers.basehttp import WSGIServer, WSGIRequestHandler
from django.core.handlers.wsgi import logger
from django.conf import settings
from django.core.management.commands import runserver
from django.utils.six.moves import socketserver
from django.utils.encoding import force_str
from .websocket import WebSocket
from .websocket_wsgi_server import WebsocketWSGIServer
from .exceptions import HandshakeError, UpgradeRequiredError
from .ws_settings import WS_SETTINGS

util._hoppish = {}.__contains__


class WebsocketRunServer(WebsocketWSGIServer):
    WS_GUID = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
    WS_VERSIONS = ('13', '8', '7')

    def upgrade_websocket(self, environ, start_response):
        """
        Attempt to upgrade the socket environ['wsgi.input'] into a websocket enabled connection.
        """
        websocket_version = environ.get('HTTP_SEC_WEBSOCKET_VERSION', '')
        if not websocket_version:
            raise UpgradeRequiredError
        elif websocket_version not in self.WS_VERSIONS:
            raise HandshakeError('Unsupported WebSocket Version: {0}'.format(websocket_version))

        key = environ.get('HTTP_SEC_WEBSOCKET_KEY', '').strip()
        if not key:
            raise HandshakeError('Sec-WebSocket-Key header is missing/empty')
        try:
            key_len = len(base64.b64decode(key))
        except TypeError:
            raise HandshakeError('Invalid key: {0}'.format(key))
        if key_len != 16:
            # 5.2.1 (3)
            raise HandshakeError('Invalid key: {0}'.format(key))

        sec_ws_accept = base64.b64encode(sha1(six.b(key) + self.WS_GUID).digest())
        if six.PY3:
            sec_ws_accept = sec_ws_accept.decode('ascii')
        headers = [
            ('Upgrade', 'websocket'),
            ('Connection', 'Upgrade'),
            ('Sec-WebSocket-Accept', sec_ws_accept),
            ('Sec-WebSocket-Version', str(websocket_version)),
        ]
        logger.debug('WebSocket request accepted, switching protocols')
        start_response(force_str('101 Switching Protocols'), headers)
        six.get_method_self(start_response).finish_content()
        return WebSocket(environ['wsgi.input'])

    def select(self, rlist, wlist, xlist, timeout=None):
        return select.select(rlist, wlist, xlist, timeout)


def run(addr, port, wsgi_handler, ipv6=False, threading=False):
    """
    manage.py runserver
    """
    server_address = (addr, port)
    if not threading:
        raise Exception("Django Websocket server must run with threading enabled")

    httpd_cls = type('WSGIServer', (socketserver.ThreadingMixIn, WSGIServer), {'daemon_threads': True})
    httpd = httpd_cls(server_address, WSGIRequestHandler, ipv6=ipv6)
    httpd.set_app(wsgi_handler)
    httpd.serve_forever()

runserver.run = run


_django_app = get_wsgi_application()
_websocket_app = WebsocketRunServer()
_websocket_url = WS_SETTINGS.get('WEBSOCKET_URL')

def application(environ, start_response):
    #todo 如果是ws链接运行_websocket_app
    if _websocket_url and environ.get('PATH_INFO').startswith(_websocket_url):
        return _websocket_app(environ, start_response)
    #todo 其他运行django默认
    return _django_app(environ, start_response)



# -*- coding: utf-8 -*-
import json
import sys
from importlib import import_module
import urlparse
from django.dispatch import Signal
from redis import StrictRedis
import django
from .ws_settings import WS_SETTINGS
from websocket_channel.publisher import RedisPublisher
from websocket_channel.subscriber import RedisSubscriber
from .web_socket_message import RedisMessage
#from .publisher import RedisPublisher
import urllib
if django.VERSION[:2] >= (1, 7):
    django.setup()
from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest, logger, STATUS_CODE_TEXT
from django.core.exceptions import PermissionDenied
from django import http
from django.utils.encoding import force_str

from django.utils.functional import SimpleLazyObject
from .exceptions import WebSocketError, HandshakeError, UpgradeRequiredError


class WebsocketWSGIServer(object):
    def __init__(self, redis_connection=None):
        """
        redis_connection can be overriden by a mock object.
        """
        self._redis_connection = redis_connection and redis_connection or StrictRedis(**WS_SETTINGS['WS_REDIS'])
        # self.Subscriber = RedisSubscriber
        self.online = 0

    def assure_protocol_requirements(self, environ):
        if environ.get('REQUEST_METHOD') != 'GET':
            raise HandshakeError('HTTP method must be a GET')

        if environ.get('SERVER_PROTOCOL') != 'HTTP/1.1':
            raise HandshakeError('HTTP server protocol must be 1.1')

        if environ.get('HTTP_UPGRADE', '').lower() != 'websocket':
            raise HandshakeError('Client does not wish to upgrade to a websocket')

    def process_request(self, request):
        request.session = None
        request.user = None
        if 'django.contrib.sessions.middleware.SessionMiddleware' in settings.MIDDLEWARE_CLASSES:
            engine = import_module(settings.SESSION_ENGINE)
            session_key = request.COOKIES.get(settings.SESSION_COOKIE_NAME, None)
            if session_key:
                request.session = engine.SessionStore(session_key)
                if 'django.contrib.auth.middleware.AuthenticationMiddleware' in settings.MIDDLEWARE_CLASSES:
                    from django.contrib.auth import get_user
                    request.user = SimpleLazyObject(lambda: get_user(request))

    def process_subscribe(self, request, recvmsg, subscriber):
        channels = urlparse.parse_qs(recvmsg).get("channel[]")
        if channels:
            comps = str(WS_SETTINGS['CHANNEL_MIDDLEWARE']).split('.')
            module = import_module('.'.join(comps[:-1]))
            [getattr(getattr(module, comps[-1])(), 'process')(request, channel, subscriber) for channel in channels]
        else:
            view = WS_SETTINGS['OTHER_REQUET']
            if view:
                comps = str(view).split('.')
                module = import_module('.'.join(comps[:-1]))
                getattr(module, comps[-1])(request, recvmsg, subscriber)

    def __call__(self, environ, start_response):
        """
        Hijack the main loop from the original thread and listen on events on the Redis
        and the Websocket filedescriptors.
        """
        websocket = None
        subscriber = RedisSubscriber(self._redis_connection)
        try:
            self.assure_protocol_requirements(environ)

            request = WSGIRequest(environ)
            self.process_request(request)
            websocket = self.upgrade_websocket(environ, start_response)
            #监听心跳消息
            subscriber.set_pubsub_channels(request, ['HEARTBEAT'])
            websocket_fd = websocket.get_file_descriptor()
            listening_fds = [websocket_fd]
            redis_fd = subscriber.get_file_descriptor()
            if redis_fd:
                listening_fds.append(redis_fd)
            subscriber.send_persited_messages(websocket)

            self.online += 1
            while websocket and not websocket.closed:
                ready = self.select(listening_fds, [], [], 4.0)[0]
                if not ready:
                    websocket.flush()
                print ready
                for fd in ready:
                    if fd == websocket_fd:
                        recvmsg = RedisMessage(websocket.receive())
                        if recvmsg:
                            self.process_subscribe(request, recvmsg, subscriber)
                    elif fd == redis_fd:
                        sendmsg = RedisMessage(subscriber.parse_response())
                        if sendmsg:
                            websocket.send(sendmsg)
                    else:
                        logger.error('Invalid file descriptor: {0}'.format(fd))
        except WebSocketError as excpt:
            logger.warning('WebSocketError: {}'.format(excpt), exc_info=sys.exc_info())
            response = http.HttpResponse(status=1001, content='Websocket Closed')
        except UpgradeRequiredError as excpt:
            logger.info('Websocket upgrade required')
            response = http.HttpResponseBadRequest(status=426, content=excpt)
        except HandshakeError as excpt:
            logger.warning('HandshakeError: {}'.format(excpt), exc_info=sys.exc_info())
            response = http.HttpResponseBadRequest(content=excpt)
        except PermissionDenied as excpt:
            logger.warning('PermissionDenied: {}'.format(excpt), exc_info=sys.exc_info())
            response = http.HttpResponseForbidden(content=excpt)
        except Exception as excpt:
            logger.error('Other Exception: {}'.format(excpt), exc_info=sys.exc_info())
            response = http.HttpResponseServerError(content=excpt)
        else:
            response = http.HttpResponse()
        finally:
            self.online -= 1
            subscriber.release()
            if websocket:
                websocket.close(code=1001, message='Websocket Closed')
            else:
                logger.warning('Starting late response on websocket')
                status_text = STATUS_CODE_TEXT.get(response.status_code, 'UNKNOWN STATUS CODE')
                status = '{0} {1}'.format(response.status_code, status_text)
                start_response(force_str(status), response._headers.values())
                logger.info('Finish non-websocket response with status code: {}'.format(response.status_code))
        return response


    def upgrade_websocket(self, environ, start_response):
        raise WebSocketError("error")


import os
from setuptools import setup, find_packages
from websocket_channel import __version__

DESCRIPTION = 'Websocket support for Django using Redis'

CLASSIFIERS = [
    'Environment :: Web Environment',
    'Framework :: Django',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Programming Language :: Python :: 2.7',
]


def read(fname):
    readme_file = os.path.join(os.path.dirname(__file__), fname)
    return os.popen('[ -x "$(which pandoc 2>/dev/null)" ] && pandoc -t rst {0} || cat {0}'.format(readme_file)).read()
setup(
    name='django-websocket-channel',
    version=__version__,
    author='Neo hu',
    author_email='9656951@qq.com',
    description=DESCRIPTION,
    long_description=read('README.md'),
    url='https://github.com/neo-hu/django-websocket-channel',
    keywords=['django', 'websocket', 'redis'],
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    include_package_data=True,
    license = "MIT",
    install_requires=[
	'wsaccel',
        'setuptools',
        'redis',
        'gevent',
        'greenlet',
        'six',
    ],
    extras_require={
        'uwsgi': ['uWSGI>=1.9.20'],
        'wsaccel': ['wsaccel>=0.6.2'],
    }
)



# entry point for the websocket loop
import os
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")
from websocket_channel.uwsgi_runserver import uWSGIWebsocketServer
application = uWSGIWebsocketServer()



#-*- coding: utf-8 -*-
from redis import ConnectionPool, StrictRedis
from .ws_settings import WS_SETTINGS
from six import string_types
from .web_socket_message import RedisStore

redis_connection_pool = ConnectionPool(**WS_SETTINGS['WS_REDIS'])


class RedisPublisher(RedisStore):
    def __init__(self, publishers):
        """
        Initialize the channels for publishing messages through the message queue.
        """
        connection = StrictRedis(connection_pool=redis_connection_pool)
        super(RedisPublisher, self).__init__(connection)
        if isinstance(publishers, string_types):
            publishers = [publishers]
        self._publishers = publishers

